import google.cloud.pubsub as pubsub
import google.gax.errors
import grpc
from gcloud.aio.pubsub.subscription import Subscription


class Topic(pubsub.topic.Topic):
    def create_if_missing(self, client=None):
        if self.exists(client=client):
            return

        try:
            self.create(client=client)
        except google.gax.errors.RetryError as e:
            if e.cause._state != grpc.StatusCode.ALREADY_EXISTS:  # pylint: disable=protected-access
                raise

    def subscription(self, name, ack_deadline=None, push_endpoint=None,
                     retain_acked_messages=None,
                     message_retention_duration=None):
        # pylint: disable=too-many-arguments
        return Subscription(
            name, self, ack_deadline=ack_deadline, push_endpoint=push_endpoint,
            retain_acked_messages=retain_acked_messages,
            message_retention_duration=message_retention_duration)



import asyncio
import json
import os
import uuid

import gcloud.aio.pubsub as pubsub


async def do_lifecycle(project, topic, subscription):
    subscriber = pubsub.Client(project)

    # create an empty topic
    topic = subscriber.topic(topic)
    topic.create_if_missing()

    # create an empty subscription
    subscription = topic.subscription(subscription)
    subscription.create_if_missing()
    subscription.pull(return_immediately=True, max_messages=1_000_000)

    # push four jobs
    num_jobs = 4

    uuids = [uuid.uuid4().hex for _ in range(num_jobs)]
    for i in range(num_jobs):
        data = {'this': {'is': {'a': {'test': uuids[i]}}}}
        topic.publish(json.dumps(data).encode('utf-8'))

    num_processed = 0
    async for job_id, message in subscription.poll():
        job = json.loads(message.data.decode('utf-8'))
        assert job['this']['is']['a']['test'] in uuids

        await subscription.acknowledge([job_id])
        num_processed += 1

        if num_processed == num_jobs:
            break

    subscription.delete()
    topic.delete()


def test_pubsub_lifecycle():
    project = os.environ['GCLOUD_PROJECT']

    topic = 'test-topic'
    subscription = 'test-subscription'

    loop = asyncio.get_event_loop()
    loop.run_until_complete(do_lifecycle(project, topic, subscription))



import google.cloud.pubsub as pubsub
from gcloud.aio.pubsub.topic import Topic


class Client(pubsub.client.Client):
    # pylint: disable=too-few-public-methods
    def topic(self, name, timestamp_messages=False):
        return Topic(name, client=self, timestamp_messages=timestamp_messages)



import asyncio
import logging

import google.auth.exceptions
import google.cloud.pubsub as pubsub
import google.gax.errors
import grpc


log = logging.getLogger(__name__)


class Subscription(pubsub.subscription.Subscription):
    async def acknowledge(self, ack_ids, client=None, retries=0):
        try:
            return super().acknowledge(ack_ids, client=client)
        except (google.auth.exceptions.TransportError,
                google.gax.errors.RetryError):
            # common intermittent communication errors
            if not retries:
                raise

            await asyncio.sleep(1)
        except Exception:
            raise

        return await self.acknowledge(ack_ids, client=client,
                                      retries=retries-1)

    def create_if_missing(self, client=None):
        if self.exists(client=client):
            return

        try:
            self.create(client=client)
        except google.gax.errors.RetryError as e:
            if e.cause._state != grpc.StatusCode.ALREADY_EXISTS:  # pylint: disable=protected-access
                raise

    async def poll(self, max_messages=1, client=None,
                   max_intermittent_errors=1, max_errors=0):
        intermittent_errors = 0
        errors = 0

        while True:
            try:
                for job in self.pull(return_immediately=False,
                                     max_messages=max_messages, client=client):
                    # reset only intermittent errors, not hard errors
                    intermittent_errors = 0

                    yield job
                    await asyncio.sleep(0)
            except (google.auth.exceptions.TransportError,
                    google.gax.errors.RetryError):
                intermittent_errors += 1
                if max_intermittent_errors is not None and \
                        intermittent_errors > max_intermittent_errors:
                    raise

                await asyncio.sleep(1)
            except Exception as e:  # pylint: disable=broad-except
                errors += 1
                if max_errors is not None and errors > max_errors:
                    raise

                log.exception(e)
                await asyncio.sleep(3)
